local VERSION = tonumber(("$Revision: 31605 $"):match("%d+"))

local PitBull = PitBull
local PitBull_WatchdogTags = PitBull:NewModule("WatchdogTags", "AceEvent-2.0")
if PitBull.revision < VERSION then
	PitBull.version = "r" .. VERSION
	PitBull.revision = VERSION
	PitBull.date = ("$Date: 2007-04-02 03:19:50 -0400 (Mon, 02 Apr 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")
end

local L = setmetatable({}, {__index = function(self, key) self[key] = key; return key end })
-- this is used as a placeholder for any localization I choose to do later.

local configMode = PitBull.configMode

local new, del = PitBull.new, PitBull.del
local wrap = PitBull.wrap

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"

local function GetClassColor(unit)
	local _,class = UnitClass(unit)
	return unpack(PitBull.colorConstants[class] or PitBull.colorConstants.unknown)
end

local function GetHostilityColor(unit)
	if UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			-- they can attack me
			if UnitCanAttack("player", unit) then
				-- and I can attack them
				return unpack(PitBull.colorConstants.hostile)
			else
				-- but I can't attack them
				return unpack(PitBull.colorConstants.civilian)
			end
		elseif UnitCanAttack("player", unit) then
			-- they can't attack me, but I can attack them
			return unpack(PitBull.colorConstants.neutral)
		elseif UnitIsPVP(unit) then
			-- on my team
			return unpack(PitBull.colorConstants.friendly)
		else
			-- either enemy or friend, no violence
			return unpack(PitBull.colorConstants.civilian)
		end
	elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) or UnitIsDead(unit) then
		return unpack(PitBull.colorConstants.tapped)
	else
		local reaction = UnitReaction(unit, "player")
		if reaction then
			if reaction >= 5 then
				return unpack(PitBull.colorConstants.friendly)
			elseif reaction == 4 then
				return unpack(PitBull.colorConstants.neutral)
			else
				return unpack(PitBull.colorConstants.hostile)
			end
		else
			return unpack(PitBull.colorConstants.unknown)
		end
	end
end

local colorNameByClass = false

local unit -- this is set before the watchdog functions are called
local healthEvents = {"UNIT_HEALTH", "UNIT_MAXHEALTH"}
local manaEvents = {"UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER"}
local partyEvents = { "PARTY_LEADER_CHANGED", "PARTY_MEMBERS_CHANGED", "RAID_ROSTER_UPDATE" }
local experienceEvents = { "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" }

local WatchdogModifiers
WatchdogModifiers = {
	-- when writing modifiers, inputs can be nil (which translates to empty string), a number, or a string. Also, any of these can be returned.
	-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
	hidezero = function(value)
		if value ~= 0 then
			return value
		end
	end,
	
	negate = function(value)
		if type(value) == "number" then
			return -value
		end
		return value
	end,
	
	percent = function(value)
		if type(value) == "number" then
			return value .. "%"
		end
		return value
	end,
	
	hidemaxhp = { function(value)
		local maxValue = UnitHealthMax(unit)
		if maxValue == 0 then
			return
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, 1, maxValue)
			if MHfound then
				if value == max then
					return
				else
					return value
				end
			end
		end
		if value == maxValue then
			return
		else
			return value
		end
	end, healthEvents },
	
	hidemaxmana = { function(value)
		if value == UnitManaMax(unit) then
			return
		end
		return value
	end, manaEvents },
	
	hidenocombat = { function(value)
		if UnitAffectingCombat(unit) then
			return value
		end
		return
	end, { "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED" } },
	
	short = function(value)
		if type(value) ~= "number" then
			if value:match("/") then
				local a,b = value:match("^(%d+)/(%d+)$")
				a = WatchdogModifiers.short(tonumber(a))
				b = WatchdogModifiers.short(tonumber(b))
				if a and b then
					return a.."/"..b
				end
			end
			return value
		end
		if value >= 10000000 or value <= -10000000 then
			return L["%.1fm"]:format(value / 1000000)
		elseif value >= 1000000 or value <= -1000000 then
			return L["%.2fm"]:format(value / 1000000)
		elseif value >= 100000 or value <= -100000 then
			return L["%dk"]:format(value / 1000)
		elseif value >= 10000 or value <= -10000 then
			return L["%.1fk"]:format(value / 1000)
		else
			return value
		end
	end,
	
	veryshort = function(value)
		if type(value) ~= "number" then
			if value:match("/") then
				local a,b = value:match("^(%d+)/(%d+)$")
				a = WatchdogModifiers.veryshort(tonumber(a))
				b = WatchdogModifiers.veryshort(tonumber(b))
				if a and b then
					return a.."/"..b
				end
			end
			return value
		end
		if value >= 1000000 or value <= -1000000 then
			return L["%dm"]:format(value / 1000000)
		elseif value >= 1000 or value <= -1000 then
			return L["%dk"]:format(value / 1000)
		else
			return value
		end
	end,
}

local WatchdogTags = {
	-- When writing tags, `unit` is modified before the call, that contains the unit id.
	-- Also, nil (which translates to empty string), a number, or a string can be returned.
	-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
	name = { function()
		local name = UnitName(unit)
		if name and colorNameByClass then
			local r, g, b = GetClassColor(unit)
			return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, name)
		end
		if not name and configMode then
			return unit
		end
		return name
	end, "UNIT_NAME_UPDATE"},
	
	namehostile = { function()
		local name = UnitName(unit)
		if not name then
			if configMode then
				return unit
			end
			return
		end
		local r, g, b
		if colorNameByClass and UnitIsPlayer(unit) then
			r, g, b = GetClassColor(unit)
		else
			r, g, b = GetHostilityColor(unit)
		end
		return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, name)
	end, "UNIT_NAME_UPDATE"},
	
	status = function()
		if UnitIsDead(unit) then
			return L["Dead"]
		elseif UnitIsGhost(unit) then
			return L["Ghost"]
		elseif not UnitIsConnected(unit) then
			return L["Offline"]
		elseif UnitAffectingCombat(unit) then
			return L["Combat"]
		elseif unit == "player" and IsResting() then
			return L["Resting"]
		end
	end,
	
	statusgone = function()
		if UnitIsDead(unit) then
			return L["Dead"]
		elseif UnitIsGhost(unit) then
			return L["Ghost"]
		elseif not UnitIsConnected(unit) then
			return L["Offline"]
		end
	end,
	
	statuscolor = function()
		if UnitIsDeadOrGhost(unit) then
			local r,g,b = unpack(defaultColors.dead)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif not UnitIsConnected(unit) then
			local r,g,b = unpack(defaultColors.disconnected)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif UnitAffectingCombat(unit) then
			local r,g,b = unpack(defaultColors.inCombat)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif u == "player" and IsResting() then
			local r,g,b = unpack(defaultColors.resting)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end,
	
	happycolor = function()
		local x = GetPetHappiness()
		if x == 3 then
			local r,g,b = unpack(defaultColors.petHappy)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif x == 2 then
			local r,g,b = unpack(defaultColors.petNeutral)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		elseif x == 1 then
			local r,g,b = unpack(defaultColors.petAngry)
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end,

	curhp = { function()
		local currValue = UnitHealth(unit)
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local maxValue = UnitHealthMax(unit)
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr
			end
		end
		return currValue or 0
	end, healthEvents },
	
	maxhp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		if maxValue == 0 then
			return 0
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue,maxValue)
			if MHfound then
				return max
			end
		end
		return maxValue
	end, healthEvents },
	
	percenthp = { function()
		local hpmax = UnitHealthMax(unit)
		if hpmax == 0 then
			return 0
		end
		return ("%.0f"):format(UnitHealth(unit) / hpmax * 100)
	end, healthEvents },
	
	missinghp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		if maxValue == 0 then
			return 0
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return max - curr
			end
		end
		return maxValue - currValue
	end, healthEvents },
	
	smarthp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		local MobHealth3 = MobHealth3
		if maxValue > 100 then
			return currValue.."/"..maxValue
		end
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr.."/"..max
			end
		end
		return ("%.0f"):format(currValue / maxValue * 100).."%" or 0
	end, healthEvents},

	curmana = { function()
		return UnitMana(unit) or 0
	end, manaEvents },
	
	maxmana = { function()
		return UnitManaMax(unit)
	end, manaEvents },
	
	percentmana = { function()
		local mpmax = UnitManaMax(unit)
		if mpmax == 0 then
			return 0
		end
		return ("%.0f"):format(UnitMana(unit) / mpmax * 100)
	end, manaEvents },
	
	missingmana = { function()
		return UnitManaMax(unit) - UnitMana(unit)
	end, manaEvents },
	
	typemana = { function()
		local p = UnitPowerType(unit)
		if p == 1 then
			return L["Rage"]
		elseif p == 2 then
			return L["Focus"]
		elseif p == 3 then
			return L["Energy"]
		else
			return L["Mana"]
		end
	end, "UNIT_DISPLAYPOWER" },
	
	level = { function()
		local x = UnitLevel(unit)
		if x > 0 then
			return x
		else
			return "??"
		end
	end, "UNIT_LEVEL" },
	
	class = function()
		return UnitClass(unit) or UNKNOWN
	end,
	
	classcolor = function()
		local r, g, b = GetClassColor(unit)
		return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end,
	
	creature = { function()
		return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
	end, "UNIT_DISPLAYPOWER" },
	
	smartclass = { function()
		if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
			return UnitClass(unit)
		end
	end, "UNIT_DISPLAYPOWER" },
	
	playerclass = { function()
		if UnitIsPlayer(unit) then
			return UnitClass(unit) or ""
		end
		return ""
	end, "UNIT_DISPLAYPOWER" },
	
	combos = { function()
		return GetComboPoints() or 0
	end, "PLAYER_COMBO_POINTS" },
	
	combosymbols = { function()
		return ("@"):rep(GetComboPoints())
	end, "PLAYER_COMBO_POINTS" },
	
	classification = function()
		if UnitClassification(unit) == "rare" then
			return L["Rare"]
		elseif UnitClassification(unit) == "rareelite" then
			return L["Rare-Elite"]
		elseif UnitClassification(unit) == "elite" then
			return L["Elite"]
		elseif UnitClassification(unit) == "worldboss" then
			return L["Boss"]
		end
	end,
	
	faction = function()
		local _, faction = UnitFactionGroup(unit)
		return faction
	end,
	
	connect = function()
		if not UnitIsConnected(unit) then
			return L["Offline"]
		end
	end,
	
	smartrace = function()
		if UnitIsPlayer(unit) then
			return UnitRace(unit) or UNKNOWN
		else
			return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
		end
	end,
	
	race = function()
		return UnitRace(unit)
	end,
	
	pvp = function()
		return UnitIsPVP(unit) and "PvP" or nil
	end,
	
	plus = function()
		return UnitIsPlusMob(unit) and "+" or nil
	end,
	
	sex = function()
		local sex = UnitSex(unit)
		if sex == 0 then
			return L["Male"]
		elseif sex == 1 then
			return L["Female"]
		end
	end,
	
	rested = function()
		if GetRestState() == 1 then
			return L["Rested"]
		end
	end,
	
	leader = { function()
		if UnitIsPartyLeader(unit) == 1 then
			return L["(L)"]
		end
	end, partyEvents },
	
	leaderlong = { function()
		if UnitIsPartyLeader(unit) == 1 then
			return L["(Leader)"]
		end
	end, partyEvents },
	
	happynum = { function()
		return GetPetHappiness() or 0
	end, "UNIT_HAPPINESS" },
	
	happytext = { function()
		return _G["PET_HAPPINESS" .. (GetPetHappiness() or 0)] or ""
	end, "UNIT_HAPPINESS" },
	
	happyicon = { function()
		local happy = GetPetHappiness()
		if happy == 3 then
			return ":D"
		elseif happy == 2 then
			return ":||"
		elseif happy == 1 then
			return "B("
		end
	end, "UNIT_HAPPINESS" },
	
	curxp = { function()
		return UnitXP(unit)
	end, experienceEvents },
	
	maxxp = { function()
		return UnitXPMax(unit)
	end, experienceEvents },
	
	percentxp = { function()
		local max = UnitXPMax(unit)
		if max > 0 then
			return ("%.0f"):format(UnitXP(unit) / max * 100)
		end
		return 0
	end, experienceEvents },
	
	missingxp = { function()
		return UnitXPMax(unit) - UnitXP(unit)
	end, experienceEvents },
	
	tappedbyme = { function()
		if UnitIsTappedByPlayer(unit) then
			return "*"
		end
	end, "Update" },

	istapped = { function()
		if UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
			return L["Tapped"]
		end
	end, "Update" },
	
	fkey = function()
		local fkey
		if u == "player" then
			fkey = 0
		else
			fkey = u:match("^party(%d)$")
		end
		if fkey then
			return "F" .. (fkey + 1)
		end
	end,
	
	white = function()
		return "|r"
	end,
	
	aggro = function()
		local val = UnitReaction("player", unit) or 5
		
		local info = UnitReactionColor[val]
		return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
	end,
	
	difficulty = function()
		if UnitCanAttack("player", unit) then
			local level = UnitLevel(unit)
			if level == 0 then
				level = 99
			end
			local info = GetDifficultyColor(level)
			return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
		end
	end,
	
	colormp = function()
		local powerType = UnitPowerType(unit)
		local r,g,b
		if powerType == 0 then
			r,g,b = unpack(defaultColors.mana)
		elseif powerType == 1 then
			r,g,b = unpack(defaultColors.rage)
		elseif powerType == 2 then
			r,g,b = unpack(defaultColors.focus)
		elseif powerType == 3 then
			r,g,b = unpack(defaultColors.energy)
		else
			r,g,b = unpack(defaultColors.unknown)
		end
		return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
	end,
	
	incombat = { function()
		if UnitAffectingCombat(unit) then
			local r,g,b = unpack(defaultColors.inCombat)
			return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		else
			return "|r"
		end
	end, { "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED" } },
	
	raidgroup = { function()
		for i = 1, GetNumRaidMembers() do
			local name, rank, subgroup = GetRaidRosterInfo(i)
			if name == UnitName(unit) then
				return "(" .. subgroup .. ")"
			end
		end
	end, "RAID_ROSTER_UPDATE" },
	
	guild = function()
		return GetGuildInfo(unit)
	end,
	
	guildrank = function()
		local _, rank = GetGuildInfo(unit)
		return rank
	end,
	
	guildbrackets = function()
		local name = GetGuildInfo(unit)
		if name then
			return "<" .. name .. ">"
		end
	end,

	target = { function()
		local u = unit .. "target"
		if UnitExists(u) then
			return UnitName(u)
		end
	end, "UNIT_TARGET" },
	
	-- aliases
	smartmissinghp = "missinghp:hidezero:negate"
}

local handler = function(literal, inner)
	local tag, modifiers = (":"):split(inner, 2)
	local func = WatchdogTags[tag]
	if func then
		local value
		if type(func) == "table" then
			value = func[1]()
		else
			value = func()
		end
		if modifiers then
			for mod in modifiers:gmatch("[^:]+") do
				if not value or value == "" then
					value = nil
				else
					local num = tonumber(value)
					if num then
						value = num
					end
				end
				local func = WatchdogModifiers[mod]
				if func then
					if type(func) == "table" then
						value = func[1](value)
					else
						value = func(value)
					end
				else
					func = WatchdogTags[mod]
					if func then
						if not value then
							if type(func) == "table" then
								value = func[1]()
							else
								value = func()
							end
						else
							return value
						end
					else
						return literal
					end
				end
			end
		end
		return value or ""
	else
		return literal
	end
end
local codeToFunction = setmetatable({}, {__index = function(self, key)
	local func = function(u)
		unit = u
		return key:gsub("(%[([a-z_:]+)%])", handler)
	end
	self[key] = func
	return func
end, __mode = 'k'})
local codeToEventList = setmetatable({}, {__index = function(self, key)
	local t = new()
	for bit in key:gmatch("%[([a-z_:]+)%]") do
		local tag, modifiers = (":"):split(bit, 2)
		tag = WatchdogTags[tag]
		if tag then
			if type(tag) == "table" then
				local event = tag[2]
				if event then
					if type(event) == "table" then
						for _,v in ipairs(event) do
							t[v] = true
						end
					else
						t[event] = true
					end
				end
			end
			if modifiers then
				for mod in modifiers:gmatch("[^:]+") do
					local modifier = WatchdogModifiers[mod] or WatchdogTags[mod]
					if type(modifier) == "table" then
						local event = modifier[2]
						if event then
							if type(event) == "table" then
								for _,v in ipairs(event) do
									t[v] = true
								end
							else
								t[event] = true
							end
						end
					end
				end
			end
		end
	end
	self[key] = t
	return t
end, __mode = 'k'})

local unalias = function(literal, inner)
	local real = "["
	local tag, modifiers = (":"):split(inner, 2)
	local func = WatchdogTags[tag]
	if func then
		if type(func) == "string" then
			real = real .. func
		else
			real = real .. tag
		end
		if modifiers then
			for mod in modifiers:gmatch("[^:]+") do
				local func = WatchdogModifiers[mod] or WatchdogTags[mod]
				if func then
					if type(func) == "string" then
						real = real .. ":" .. func
					else
						real = real .. ":" .. mod
					end
				else
					return literal
				end
			end
		end
		return real .. "]"
	else
		return literal
	end
end
local antiAlias = setmetatable({}, {__index = function(self, key)
	local value = key:gsub("(%[([a-z_:]+)%])", unalias)
	self[key] = value
	return value
end, __mode= 'k'})

local eventData = {}

local watchdogRegistry = {}
PitBull.watchdogRegistry = watchdogRegistry

local codeHash = {}
local unitHash = {}

local function scheduleFunc()
	for text in pairs(eventData[event]) do
		if text:IsShown() then
			text:SetText(codeToFunction[codeHash[text]](unitHash[text]))
		end
	end
end
local function eventFunc(units, event)
	local checkUnits = next(units)
	for text in pairs(eventData[event]) do
		if text:IsShown() then
			local unit = unitHash[text]
			if not checkUnits or units[unit] then
				text:SetText(codeToFunction[codeHash[text]](unit))
			end
		end
	end
end
local IsWackyUnit = PitBull.IsWackyUnit
function PitBull:RegisterWatchdogFontString(text, unit, code)
	if code == nil then
		code = ""
	end
	code = antiAlias[code]
	if codeHash[text] then
		if codeHash[text] == code and unitHash[text] == unit then
			if text:IsShown() then
				text:SetText(codeToFunction[codeHash[text]](unitHash[text]))
			end
			return
		end
		self:UnregisterWatchdogFontString(text)
	end
	codeHash[text] = code
	unitHash[text] = unit
	if not IsWackyUnit[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			if not t then
				eventData[event] = new()
				t = eventData[event]
			
				if event == "Update" then
					PitBull_WatchdogTags:ScheduleRepeatingEvent("PitBull-WatchdogTagUpdater", func, 0.15, scheduleFunc)
				else
					PitBull_WatchdogTags:RegisterBucketEvent(event, 0.05, eventFunc, event)
				end
			end
			t[text] = true
		end
	end
	watchdogRegistry[text] = unit
	
	if text:IsShown() then
		text:SetText(codeToFunction[code](unit))
	end
end
PitBull.RegisterWatchdogFontString = wrap(PitBull.RegisterWatchdogFontString, ':RegisterWatchdogFontString')

function PitBull:UnregisterWatchdogFontString(text)
	local code = codeHash[text]
	if not code then
		return
	end
	
	watchdogRegistry[text] = nil
	codeHash[text] = nil
	local unit = unitHash[text]
	unitHash[text] = nil
	if not IsWackyUnit[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			assert(t)
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event == "Update" then
					PitBull_WatchdogTags:CancelScheduledEvent("PitBull-WatchdogTagUpdater")
				else
					PitBull_WatchdogTags:UnregisterBucketEvent(event)
				end
			end
		end
	end
	
	text:SetText("")
end
PitBull.UnregisterWatchdogFontString = wrap(PitBull.UnregisterWatchdogFontString, ':UnregisterWatchdogFontString')

function PitBull:UpdateWatchdogFontString(text)
	if watchdogRegistry[text] and text:IsShown() then
		text:SetText(codeToFunction[codeHash[text]](unitHash[text]))
	end
end
PitBull.UpdateWatchdogFontString = wrap(PitBull.UpdateWatchdogFontString, ':UpdateWatchdogFontString')

function PitBull_WatchdogTags:OnChangeConfigMode(value)
	configMode = value
	for text in pairs(watchdogRegistry) do
		PitBull:UpdateWatchdogFontString(text)
	end
end